( * 3iBeats.fs - AE )
( This software module runs on 3iForth.avr or 3iSensor.avr if using matrix keypad.  It does not require 3iBeats.avr. )

14 constant buzzer

\ Potentiometer pins are 15/A1 (out) -> A7 (analog in) + pot (A10K pot - audio A taper pot has better performance in low range)
15 constant potout
21 constant potin

\ mimicking 3iForth.fs, parsing bits of a hex word
: .star ( n -- ) 0> if '*' else bl then emit ;  \ star and blank
: .note ( n -- ) 0> if 14  else 22 then emit ;  \ note and rest
: .stars (  bitfield  --  )  \  loop through bits from MSB mask and print  \ used in 3iFont.fs
    init-mask begin 2dup and .star 1 rshift dup 0= until 2drop ;
: .notes (  bitfield  --  )  \  loop through bits from MSB mask and print   \ used in 3iBeat.fs and 3iMenu.fs
    init-mask begin 2dup and .note 1 rshift dup 0= until 2drop ;

: nnotes ( n-elements N --  n-elements ) dup 0 do dup i - pick .notes loop drop ; \ needed in 3iShift.fs beats sequencer

\ APPLICATION
variable beatdur
: beatdur! ( dur -- ) beatdur ! ;
: beatdur@ ( -- dur ) beatdur @ ;  \ TODO: have this be read by input POTENTIOMETER for rate control

: init-beats
  buzzer out
  100 beatdur!  \ default - 16th note at 100 ms
  potout out
  potout lo
; 

: readpot ( -- ) potout hi potin pinreadLevel 2 / dup beatdur@ = if drop else dup . beatdur! then ;  \ sets the 16th note beatdur from 0 to 200 ms  ( divider = 5 )
: setpot ( -- ) begin readpot key? until ;

\ EXPERIMENTAL - 2021-05-31---Mon---23:33 - tones instead of beats
: d___. d___ freq lohi 20 lohi buzzer [tone] ;
: d_.   d_ freq lohi 20 lohi buzzer [tone] ;

variable npercustones
create percustones ' d___. , ' d_. , ' d_. , ' d_. ,
variable tonecount
: init-percustones ( numtones -- )
    npercustones ! 
    0 tonecount ! ;
: pulse-tone ( -- )   \  each pulse plays one of the percussion tones in circular buffer
    tonecount @ dup 
    4 * percustones + @ execute  \ get current tone 
    1 + npercustones @ mod tonecount !  \ save next tone
    ;

: // beatdur@ ms ;  \ Timings from Forth, 100ms = 16th note, 200ms = 8th note, 400ms = quarter note, and 800ms = half note, 1600ms = whole note
: metronome  begin // ( buzzer pulse-hi )  pulse-tone readpot key? until ;  \ metronome
: Nmetronome ( dur -- ) >r begin r@ ms ( buzzer pulse-hi ) pulse-tone key? until r> drop ; 



\ beats: 237, 118, 17, 13, 213, 253, 0
\ seqs: 0 213 13 17    and   1 253 13 17 0 213 13 17
: beat-byte-lsb  ( b -- ) \ parse byte to beat - LSB first for shift in register
    8 0 do dup 1 and if ( buzzer pulse-hi ) pulse-tone then // 1 rshift loop drop ;  
: Nbeat-lsb ( N-bytes N -- ) 0 do beat-byte-lsb loop ;  \  237 118  2 Nbeat
: Nseq-lsb ( N-bytes N -- N-bytes ) >r begin r@ ndup r@ Nbeat-lsb readpot key? until r> drop ;
: demo-uptown-funk 100 beatdur! 1 253 13 17 0 213 13 17   8 Nseq-lsb ;  \ 16 notes = 4 bars uptown funky bunch: https://www.youtube.com/watch?v=M1F0lBnsnkE

: beat-byte-msb  ( b -- )  \ parse byte to beat - MSB first for keyboard entry
    9 1 do dup 1  8 i - lshift  and  8 i - rshift if ( buzzer pulse-hi ) pulse-tone then // loop drop ;
: Nbeat-msb ( N-bytes N -- N-bytes ) dup 0 do dup i - pick beat-byte-msb loop drop ;  \  156 170  2 Nbeat-msb
: Nseq-msb ( N-bytes N -- N-bytes ) >r r@ nnotes begin r@ Nbeat-msb ( readpot ) key? until r> drop ;


\ APPLICATION IDEA: Use a single 8-pin DIP to create 2 note combinations, and OLED spits out their binary encoding, 
\ But touchpad entry allows typing the beats in directly, and uC accepts them as the basis for looping.
\ THIS IS WHERE SOFTWARE BEATS HARDWARE!
\ With a mixed interface (software + hardware) we get the best of both worlds - optimized user interaction/interface, and precise timing independent of Microsoft Windows, and not requiring a computer to run.

\  KEYPAD TEST -- include 3iSensor.fs

: beep ( -- ) buzzer hi 5 ms buzzer lo 50 ms ;
: buzzcount ( N -- ) 0 do beep loop ;
: keypad-demo ( -- ) 
    begin  fastkeypad 1 + buzzcount  key?  until  ;

create beatspad 0 c, 1 c, 0 c, 0 c, 0 c, 0 c, 0 c, 0 c, 0 c, 0 c, 0 c, 0 c, 0 c, 0 c, 0 c, 0 c, 
        
\ APPLICATION - use 74165 to allow a simple ATTINY to monitor a keypad continuously and send out the results, so not to slow down the main uC.  Another application for the GA144...

: drawnote ( ascii -- ) dup 1 = if 14 else 22 then emit space ;
: nibblespace? ( -- ) 1 + 4 mod 0 = if 2 spaces then ;  \ put space after each 4 bits (nibble) to aid visual perception
: beats-demo ( -- )
cr ." Enter 2 bars of 8x 16th-note beats (press note or rest): " cr
    8 0 do fastkeypad bytes beatspad + c@ drawnote i nibblespace? loop >b8 1 Nseq-msb  \ get 8 entries, then loop it
; 

\ TODO: Expand to a menu system:  ADD bar, EDIT, DELETE bar.  
\       Extract from keypad to fewest number of buttons
\       Add potentiometer for rate control
\       Incorporate keypad with 74165 shift-in monitor to allow controlling keypad with 3 input wires instead of 8
\       Incorporate OLED
\       Incorporate full menu system: Add bar, delete bar, fwd/back.   5 buttons   - < o > + 
\       Trigger keypad from interrupt to avoid need to continually monitor.
\       Run in code .avr, move to Attiny84 (11 pins) for dedicated keypad monitoring... but need Attiny comms to Nano

\ ======================================================================

: help-beats
  cr
  cr ." BEATS (help-beats)"
  cr 
  cr ." hardware: requires Nano, either 74165 IC on nano pins 8,9,10, 8-bit dip or switches, buzzer on nano pin 4"
  cr ."         OR  Nano, 4x4 matrix keypad on nano pins 2-9, buzzer on nano pin 17"
  cr ." include   3iForth.fs  3iShift.fs  3iSensor.fs 3iBeats.fs   (requires Nano running 3iForth.avr)"
  cr ." 1 open-com   rx"
  cr
  cr ." init-shiftin  OR  init-beats"
  cr ." 200 Nmetronome (quarter notes)  metronome (16th notes)"
  cr ." latch-in [sample]  or  N Nsample   N Nsamplebeat"
  cr
  cr ." keypad-init   create keypad (16 values)"
  cr ." fastkeypad   fastkeypad-test"
  cr ." N Nbeat  N Nseq-msb  N Nseq-lsb"
  cr ." demo-uptown-funk"
  cr ." beats-demo"
  cr
;

help-beats

\ V2 - 2021-05-26---Wed---07:52  extracted from Shift.fs.  Added keypad entry with software prompt for 2 bars
\ V3 - 2021-06-01---Tue---02:04  added: beat sequence with tones -- for two-tone drum for e.g. garbi -- bass and snare or clap & snap
\     Design: put a speaker with musical tones to simulate deep and high, e.g. bass and timpani drums etc.
\     Then the design cycles through executables of tone, matched to the positive beats, instead of just to pulse-hi
\     This is where ColorForth is better because it adds a few register elements to hold indexed elements/additional stack
\     In this case, e.g. there could be a stack with just these elements.  That's what memory is... 
\     the advantage in ColorForth is use of a registers instead of memory for this, i.e. it is much faster then
