( 3iI2C.fs - 2021-06-03---Thu---14:24 )
( GForth file for communicating directly with OLED display's SSD1306 driver IC, allowing for bare metal development, with use of e.g. Adafruit's SSD1306 device driver software or their GFX graphics library. )

( 2021-06-08---Tue---13:13 - revisiting following completion of 3iI2C.avr )


: i2c-init ( -- ) \ setup Wire instance
    CALL_I2C_INIT xcall 
    $3C 31 setreg  \ store address
;

: i2c-setclock ( val/10k -- ) \ e.g. 10 means 100kHz bitrate
    xstore CALL_I2C_SETCLOCK xcall ;

: i2c-restoreclock ( val/10k -- ) \ e.g. 10 means 100kHz bitrate
    xstore CALL_I2C_SETCLOCK xcall ;

: [oled-bytes]  ( b1 b2 b3 ... bn N -- )  \ sends N bytes in reverse order and puts N at the end (target TOS)
    >r r@ revstore r> xstore ;

: oled-cmd  [oled-bytes] CALL_I2C_CMDOLED xcall ;

: oled-data [oled-bytes] CALL_I2C_DATAOLED xcall ;

\ ===================================================

\ display on/off
: oled-sleep $AE 1 oled-cmd ;  \ default
: oled-wake  $AF 1 oled-cmd ;

\ charge pump
: oled-bright $8D $14 2 oled-cmd ; \ enable
: oled-dim    $8D $10 2 oled-cmd ; \ disable (default)

\ contrast control
: oled-contrast  $81 $8F 2 oled-cmd ; \ contrast control 256-steps ($FF brightest, $0 dimmest), brighter consumes more current

\ display mode
: oled-normal  $A6 1 oled-cmd ;
: oled-invert  $A7 1 oled-cmd ;

\ memory address mode
: oled-col+     $20 $00 2 oled-cmd ;  \ horizontal increment - auto-advance to right, then down, the re-start at top left
: oled-row+     $20 $01 2 oled-cmd ;  \ vertical increment - auto-advance down, then to right
: oled-page+    $20 $10 2 oled-cmd ;  \ auto-advance to right, but looping in same row (page)  default

\ scanning setup - allows installing the display in any layout and setting the top in software.  Also, can do screen flip on the fly
: [oled-pins-top]
  $C0 $8 or     \ rowscan direction  COMSCAN INC/DEC  $8 or $0
  $A0 $1 or     \ colscan direction  SEGRE (segment remap)  $1 or $0
  2 oled-cmd ;
: oled-pins-top  [oled-pins-top] oled-col+ ;  \ for l2r scanning in this orientation

: [oled-pins-bottom]
  $C0 $0 or     \ rowscan direction  COMSCAN INC/DEC  $8 or $0
  $A0 $0 or     \ colscan direction  SEGRE (segment remap)  $1 or $0
  2 oled-cmd ;
: oled-pins-bottom [oled-pins-bottom] oled-col+ ; \ for l2r scanning in this orientation

\ restricts pixel on to window range: rows = [0,7], column = [0,$7F], off is unchanged, existing content is unchanged
: oled-window-full ( -- )  $22 0 7 $21 0 $7F    6 oled-cmd ;   \ full screen
: oled-rowstart ( -- ) oled-page+ $10 $00 2 oled-cmd oled-col+ ; \ Reset col to 0 (of 127) at start of current row
: oled-colloc ( n -- ) $10 /mod oled-page+ $10 + swap $00 + 2 oled-cmd oled-col+ ;
: set-pixel ( -- ) $1 1 oled-data ; \ D0 top of the 8 pixel colbyte
\ TODO - read colbyte GDDRAM memory so can mask the additional pixel on/off or not (vs. create/save display buffer)

  
: oled-fill-lock $A5 1 oled-cmd ;  \ hardware instruction (clears GDDRAM) but locks the display until 
: oled-fill-unlock $A4 1 oled-cmd ;
  
: oled-rowchar ( asc -- )  16 0 do dup loop drop   16 oled-data ;   
: oled-clearrow  $0    oled-rowchar  ;
: oled-fillrow   $FF  oled-rowchar  ;
: oled-clear8  8 0 do oled-clearrow loop ;
: oled-fill8  8 0 do oled-fillrow loop ;
: [oled-clear]  oled-row+ 8 0 do oled-clear8 loop oled-col+ ;
: [oled-fill]   oled-row+ 8 0 do oled-fill8  loop oled-col+ ;
: oled-clear  oled-window-full [oled-clear]  ;
: oled-fill   oled-window-full [oled-fill]   ;

\ FEATURE: store pins top/bottom state, and clear screen by flipping memory mode row+/col+ to get a nice smooth scrool
\ may not be as relevant when clearing directly from uC (faster)

: oled-num8  s" 12345678" ..oled ;
: oled-num10 s" 0123456789" ..oled ;
: oled-numrow oled-num8 oled-num8 ;
: oled-num8fill 8 0 do oled-numrow loop ;
: oled-num8fill 8 0 do oled-numrow loop ;
: oled-num8fill 8 0 do oled-numrow loop ;
: oled-num10fill 13 0 do oled-num10 loop ;

\ can fill 16x8 char (total 128) in a 128x64 pixel screen  (each character is 8x8 pixel block)
variable count 
: count! ( n -- ) count c! ;
: count@ ( -- n ) count c@ ;
: count++ ( -- c' ) count@ 1 + 10 mod dup count! ;
\ Fills screen with continuously increasing text --- good for testing scrolling functionality
: oled-textroll count@ 128 0 do '0' + .oled count++ loop drop ;


: demo-flip
  \ pins top
  oled-pins-top  oled-window-full oled-clear 
  s" Hello World!" ..oled
  key drop
  \ pins down
  oled-pins-bottom oled-window-full oled-clear 
  s" Hello World!" ..oled
;

\ TODO: DEMONSTRATE writing in row+ mode,  unrotated and rotated

: oled-init
  i2c-init
  10 i2c-setclock
  oled-bright
  oled-contrast
  oled-pins-top
  oled-window-full
  oled-fill
  oled-wake
;


: oled-init1 ( -- )
    $D5 $80  \ osc. frequency controlling frame frequency
    $A8 63   \ set rows (multiplex)
    $DA $12  \ set COM pins configuration
    $D9 $F1  \ set per-charge period
    8 oled-cmd ;
    
: oled-init2 ( -- )
    $D3 $0    \ start row (offset)
    $40 $0 or \ memory map start row
    3 oled-cmd ;

: oled-init6 ( -- )
    $DB $40 $A4 $A6 $2E $AF  6 oled-cmd ;

   
 

\ ====================================

: help-i2c
  cr
  cr ." Help I2C (help-i2c)"
  cr
  cr ." include 3iI2C.fs"
  cr ." include 3iFont.fs for text"
  cr
  cr ." i2c-init   10 i2c-setclock"
  cr ." oled-wake  oled-sleep"
  cr ." oled-contrast"
  cr ." oled-bright  oled-dim"
  cr ." oled-col+  oled-pins-top"
  cr
  cr ." oled-fill  oled-clear" 
  cr
  cr ." oled-init  ALL OF ABOvE "
  cr
;



\ need I2C communication primitives (see Wire.h/.cpp)
\ Excellent tutorial: https://mansfield-devine.com/speculatrix/2018/02/avr-basics-using-the-i2c-bus-2-transmitting/
\ primitive is Arduino's setClock() function which sets clock speed / crystal
\ safe default for I2C is 100 kHz (standard speed) or 400 kHz (fast).  Nano clock speed is 16MHz.
\ CPU frequency is scaled so it cannot generate all possible frequencies

\ : setclk
\    TWBR = ((F_CPU / frequency) - 16) / 2;  \ sets register  .e.g. 16000 / 100 ->72  - note this uses TWI (two-wire interface)

\ ==========================================

( Writing this by translating the driver software into Forth.  )
( There is so much to discard in C! C code feels like obfuscation-ware )

\ V1 - 2021-06-09---Wed---04:27  - WORKS! (with revised 3iForth.fs comms protocol

\ Next Steps:
\   + Embedded turtle logo Forth!
\   + Joystick or key or button control with an OLED I2C display
\   + If use TFT screen can actually do it with a TFT display
\   + positioning cursor arbitrarily > row+ writing rotated/unrotated.  Pixel activation...
\   + lf and cr (newline)
\   + easy display scrolling

\   + Graphics: how to set the ROW?  
\   + How to read the RAM to mask (set pixel) without clobbering other pixels?
\       Maybe it's to set the address to the READ mode, i.e. 0x3D instead of 3C?
\   http://robotcantalk.blogspot.com/2015/03/interfacing-arduino-with-ssd1306-driven.html
\   + Scrolling screen...
\ EXCELLENT DESCRIPTION: https://iotexpert.com/debugging-ssd1306-display-problems/

\ I2C Scanner code: https://wiki.seeedstudio.com/Arduino_Software_I2C_user_guide/
\ I2C address is 7-bits, then left-shifted for R/W bit...  Each OLED SSD1306 driver has two addresses, one for read/write...
\    https://www.ccsinfo.com/forum/viewtopic.php?t=52836

\   REFLASH THE FIRMWARE - SIMPLIFIED THE I2C OLED INTERFACE BY REMOVING RESTORECLOCK CMD (forth.avr, i2c.avr, forth.fs, i2c.fs)

\   + Graphics - Piano visual display simulator
\   + Graphics - Pong the game
\   + Addressible LEDs: - https://www.thesmarthomehookup.com/the-complete-guide-to-selecting-individually-addressable-led-strips/


\ include 3iI2C.fs
\ include 3iFont.fs
\ include 3iI2C.fs
\ include 3iArt.fs
\ include 3iArt-butterfly.fs
\ include 3iArt-portraits.fs
